package com.lambdazen.bitsy;

import java.util.Collections;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.UUID;
import java.util.concurrent.CopyOnWriteArraySet;

import com.tinkerpop.blueprints.Element;

public abstract class BitsyElement implements Element {
    UUID id;
    Map<String, Object> properties;
    ITransaction tx;
    BitsyState state;
    int version;
    boolean updated;
    
    public BitsyElement(UUID id, Map<String, Object> properties, ITransaction tx, BitsyState state, int version) {
        this.id = id;
        this.properties = properties;
        this.tx = tx;
        this.state = state;
        this.version = version;
        this.updated = false;
    }
    
    public Object getId() {
        // No TX check to return the ID
        
        return id;
    }
    
    public <T>T getProperty(String key) {
        tx.validateForQuery(this);
        
        if (properties == null) {
            return null;
        } else {
            return (T) (properties.get(key));
        }
    }

    public Set<String> getPropertyKeys() {
        tx.validateForQuery(this);
        
        if (properties == null) {
            return Collections.emptySet();
        } else {
            return new CopyOnWriteArraySet<String>(properties.keySet());
        }
    }
    
    public <T>T removeProperty(String key) {
        markForUpdate();
        
        assert (properties != null);

        return (T) (properties.remove(key));
    }

    public void setProperty(String key, Object value) {
        if (value == null) {
            throw new IllegalArgumentException("A null property can not be stored. You can call removeProperty() instead");
        }

        markForUpdate();
        
        assert (properties != null);
        
        if (key == null) {
            throw new IllegalArgumentException("Expecting non-null key in setProperty");
        } else if (key.length() == 0) {
            throw new IllegalArgumentException("Expecting non-empty key in setProperty");
        } else if (key.equals("id")) {
            throw new IllegalArgumentException("Can not set the 'id' property on an element");
        } else if (key.equals("label")) {
            throw new IllegalArgumentException("Can not set the 'label' property on an element");
        }
        
        properties.put(key, value);
    }
    
    /** This method prepares the vertex/edge for an update */
    public void markForUpdate() {
        if (!updated) {
            updated = true;
            
            // Make a copy of the underlying property map
            if (properties != null) {
                properties = new TreeMap<String, Object>(properties);
            } else {
                properties = new TreeMap<String, Object>();
            }
        }

        tx.markForPropertyUpdate(this);
    }
    
    /** Added in Blueprints 2.3.0 to remove underlying vertex or edge */
    public abstract void remove();
    
    public BitsyState getState() {
        return state;
    }
    
    public void setState(BitsyState state) {
        this.state = state;
    }

    public int getVersion() {
        return version;
    }

    public ITransaction getTransaction() {
        return tx;
    }
    
    @Override
    public boolean equals(Object o) {
        if (o instanceof BitsyElement) {
            return ((BitsyElement)o).getId().equals(getId());
        } else {
            return false;
        }
    }

    @Override
    public int hashCode() {
        return getId().hashCode();
    }
}