package com.lambdazen.bitsy.store;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import com.lambdazen.bitsy.BitsyEdge;
import com.lambdazen.bitsy.BitsyErrorCodes;
import com.lambdazen.bitsy.BitsyException;
import com.lambdazen.bitsy.BitsyRetryException;
import com.lambdazen.bitsy.BitsyVertex;
import com.lambdazen.bitsy.ICommitChanges;
import com.lambdazen.bitsy.IEdge;
import com.lambdazen.bitsy.IGraphStore;
import com.lambdazen.bitsy.index.EdgeIndexMap;
import com.lambdazen.bitsy.index.VertexIndexMap;
import com.tinkerpop.blueprints.Direction;
import com.tinkerpop.blueprints.Edge;
import com.tinkerpop.blueprints.Element;
import com.tinkerpop.blueprints.Vertex;

/**
 * This class implements a MapDB-backed store for a graph, along with its key
 * indexes.
 */
public class MemoryGraphStore implements IGraphStore {
//    private static final Logger log = LoggerFactory.getLogger(MemoryGraphStore.class);

    private ReadWriteLock rwLock;
    private Map<UUID, VertexBean> vertices;
    private Map<UUID, EdgeBean> edges;
    private AdjacencyMap adjMap;
    private VertexIndexMap vIndexMap;
    private EdgeIndexMap eIndexMap;
    private boolean allowFullGraphScans;

    public MemoryGraphStore(boolean allowFullGraphScans) {
        this.rwLock = new ReentrantReadWriteLock(true);
        this.allowFullGraphScans = allowFullGraphScans;
        
        reset();
    }

    protected void reset() {
        this.vertices = new ConcurrentHashMap<UUID, VertexBean>();
        this.edges = new ConcurrentHashMap<UUID, EdgeBean>();

        this.adjMap = new AdjacencyMap(false, new IEdgeRemover() {
            @Override
            public IEdge removeEdge(UUID id) {
                return edges.remove(id);
            }
        });
        this.vIndexMap = new VertexIndexMap();
        this.eIndexMap = new EdgeIndexMap();
    }

    @Override
    public boolean allowFullGraphScans() {
        return allowFullGraphScans;
    }

    /**
     * This method commits a set of changes. It requires a write lock on the Map
     * DB.
     */
    public void commit(ICommitChanges changes) {
        commit(changes, true);
    }

    // This method is called with incrementVersions=false from
    // FileBackedMemoryGraphStore
    public void commit(ICommitChanges changes, boolean incrementVersions) {
        rwLock.writeLock().lock();

        try {
            checkForConcurrentModifications(changes, incrementVersions);

            saveChanges(changes);
        } finally {
            rwLock.writeLock().unlock();
        }
    }

    private void checkForConcurrentModifications(ICommitChanges changes, boolean incrementVersions) {
        // Check the versions
        for (BitsyVertex vertex : changes.getVertexChanges()) {
            if (incrementVersions) {
                // Increment the version before the commit
                vertex.incrementVersion();
            }

            UUID key = (UUID) vertex.getId();

            switch (vertex.getState()) {
            case U:
                break;

            case D:
            case M:
                VertexBean vb = vertices.get(key);
                if (vb != null && (vb.getVersion() + 1 != vertex.getVersion())) {
                    throw new BitsyRetryException(BitsyErrorCodes.CONCURRENT_MODIFICATION, "Vertex " + key + " was modified. Loaded version "
                            + (vertex.getVersion() - 1) + ". Current version in DB: " + vb.getVersion());
                }
            }
        }

        // Check the versions of edge next
        for (BitsyEdge edge : changes.getEdgeChanges()) {
            if (incrementVersions) {
                // Increment the version before the commit
                edge.incrementVersion();
            }

            UUID key = (UUID) edge.getId();

            switch (edge.getState()) {
            case U:
                break;

            case D:
            case M:
                EdgeBean eb = edges.get(key);
                if (eb != null && (eb.getVersion() + 1) != edge.getVersion()) {
                    throw new BitsyRetryException(BitsyErrorCodes.CONCURRENT_MODIFICATION, "Edge " + key + " was modified. Loaded version "
                            + (edge.getVersion() - 1) + ". Current version in DB: " + eb.getVersion());
                }
            }
        }
    }

    protected long saveChanges(ICommitChanges changes) {
        return saveChanges(changes, null);
    }
    
    // This method is used by commit (with increment option) and the initial
    // load from DB (without increment option)
    protected long saveChanges(ICommitChanges changes, IStringCanonicalizer canonicalizer) {
        long addedVE = 0;

        // All OK. Update vertex
        for (BitsyVertex vertex : changes.getVertexChanges()) {
            addedVE = saveVertex(addedVE, vertex, canonicalizer);
        }

        // Process the edges next
        for (BitsyEdge edge : changes.getEdgeChanges()) {
            addedVE = saveEdge(addedVE, edge, canonicalizer);
        }

        return addedVE;
    }

    protected long saveEdge(long addedVE, BitsyEdge edge, IStringCanonicalizer canonicalizer) {
        UUID key = (UUID) edge.getId();

        switch (edge.getState()) {
        case U:
            break;

        case D:
            eIndexMap.remove(edges.get(key));
            //log.debug("Removing edge {}", edge.getId());

            // Remove this edge from incoming and outgoing vertices
            adjMap.removeEdge(key, edge.getOutVertexId(), edge.getLabel(), edge.getInVertexId());
            edges.remove(key);
            addedVE--;

            break;

        case M:
            EdgeBean eBean = (canonicalizer == null) ? edge.asBean() : edge.asBean(canonicalizer);
            // TODO: Pick up UUIDs for use in EdgeBean 
            if (!vertices.containsKey(edge.getInVertexId()) || !vertices.containsKey(edge.getOutVertexId())) {
                //log.debug("Skipping edge {}", edge.getId());
            } else {
                //log.debug("Modifying edge {}", edge.getId());
                eIndexMap.remove(edges.get(key));
                eIndexMap.add(eBean);

                if (edges.put(key, eBean) == null) {
                    addedVE++;
                }

                adjMap.addEdge(key, edge.getOutVertexId(), edge.getLabel(), edge.getInVertexId(), edge.getVersion());
            }
        }
        return addedVE;
    }

    protected long saveVertex(long addedVE, BitsyVertex vertex, IStringCanonicalizer canonicalizer) {
        UUID key = (UUID) vertex.getId();

        switch (vertex.getState()) {
        case U:
            break;

        case D:
            //log.debug("Deleting vertex {}", key);
            vIndexMap.remove(vertices.get(key));
            adjMap.removeVertex(key);
            vertices.remove(key);
            addedVE--;

            break;

        case M:
            //log.debug("Updating vertex {}", key);
            VertexBean vBean = (canonicalizer == null) ? vertex.asBean() : vertex.asBean(canonicalizer);
            vIndexMap.remove(vertices.get(key));
            vIndexMap.add(vBean);
            if (vertices.put(key, vBean) == null) {
                addedVE++;
            }
        }
        return addedVE;
    }

    public VertexBean getVertex(UUID id) {
        // Not using a read lock because the ID is available 
        return vertices.get(id);
    }

    public EdgeBean getEdge(UUID id) {
        // Not using a read lock because the ID is available 
        return edges.get(id);
    }

    // Methods that only need a read lock
    public List<EdgeBean> getEdges(UUID vertexId, Direction dir, String[] edgeLabels) {
        rwLock.readLock().lock();

        try {
            List<UUID> edgeIds = adjMap.getEdges(vertexId, dir, edgeLabels);
            List<EdgeBean> ans = new ArrayList<EdgeBean>(edgeIds.size());

            for (UUID edgeId : edgeIds) {
                EdgeBean edge = getEdge(edgeId);

                assert edge != null;

                ans.add(edge);
            }

            return ans;
        } finally {
            rwLock.readLock().unlock();
        }
    }

    @Override
    public Collection<VertexBean> getAllVertices() {
        // This method exposes the underlying collection, without a read/write
        // lock. The idea is to not block other operations while a reader is
        // cycling through all the vertices. There returned list may include
        // some vertices from ongoing transaction, but not others.

        return vertices.values();
    }

    @Override
    public Collection<EdgeBean> getAllEdges() {
        return edges.values();
    }

    @Override
    public <T extends Element> void createKeyIndex(String key, Class<T> elementType) {
        rwLock.writeLock().lock();

        try {
            if (elementType.equals(Vertex.class)) {
                vIndexMap.createKeyIndex(key, getAllVertices().iterator());
            } else if (elementType.equals(Edge.class)) {
                eIndexMap.createKeyIndex(key, getAllEdges().iterator());
            } else {
                throw new BitsyException(BitsyErrorCodes.UNSUPPORTED_INDEX_TYPE, "Encountered index type: " + elementType);
            }
        } finally {
            rwLock.writeLock().unlock();
        }
    }

    @Override
    public <T extends Element> void dropKeyIndex(String key, Class<T> elementType) {
        rwLock.writeLock().lock();

        try {
            if (elementType.equals(Vertex.class)) {
                vIndexMap.dropKeyIndex(key);
            } else if (elementType.equals(Edge.class)) {
                eIndexMap.dropKeyIndex(key);
            } else {
                throw new BitsyException(BitsyErrorCodes.UNSUPPORTED_INDEX_TYPE, "Encountered index type: " + elementType);
            }
        } finally {
            rwLock.writeLock().unlock();
        }
    }

    @Override
    public <T extends Element> Set<String> getIndexedKeys(Class<T> elementType) {
        rwLock.readLock().lock();

        try {
            if (elementType.equals(Vertex.class)) {
                return vIndexMap.getIndexedKeys();
            } else if (elementType.equals(Edge.class)) {
                return eIndexMap.getIndexedKeys();
            } else {
                throw new BitsyException(BitsyErrorCodes.UNSUPPORTED_INDEX_TYPE, "Encountered index type: " + elementType);
            }
        } finally {
            rwLock.readLock().unlock();
        }
    }

    @Override
    public Collection<VertexBean> lookupVertices(String key, Object value) {
        rwLock.readLock().lock();

        try {
            return vIndexMap.get(key, value);
        } finally {
            rwLock.readLock().unlock();
        }
    }

    @Override
    public Collection<EdgeBean> lookupEdges(String key, Object value) {
        rwLock.readLock().lock();

        try {
            return eIndexMap.get(key, value);
        } finally {
            rwLock.readLock().unlock();
        }
    }

    protected AdjacencyMap getAdjacencyMap() {
        return adjMap;
    }

    @Override
    public void shutdown() {
        reset();
    }
}
